# Copyright 2018 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ----------------------------------------------------------------------------
#
#     ***     AUTO GENERATED CODE    ***    AUTO GENERATED CODE     ***
#
# ----------------------------------------------------------------------------
#
#     This file is automatically generated by Magic Modules and manual
#     changes will be clobbered when the file is regenerated.
#
#     Please read more about how to change this file in README.md and
#     CONTRIBUTING.md located at the root of this package.
#
# ----------------------------------------------------------------------------

# Add our google/ lib
$LOAD_PATH.unshift ::File.expand_path('../libraries', ::File.dirname(__FILE__))

require 'chef/resource'
require 'google/hash_utils'
require 'google/sql/network/get'
require 'google/sql/property/boolean'
require 'google/sql/property/integer'
require 'google/sql/property/string'
require 'google/sql/property/string_array'

module Google
  module GSQL
    # A provider to manage Google Cloud SQL resources.
    # rubocop:disable Metrics/ClassLength
    class Flag < Chef::Resource
      resource_name :gsql_flag

      # allowed_string_values is Array of Google::Sql::Property::StringArray
      property :allowed_string_values,
               Array, coerce: ::Google::Sql::Property::StringArray.coerce, desired_state: true
      # applies_to is Array of Google::Sql::Property::StringArray
      property :applies_to,
               Array, coerce: ::Google::Sql::Property::StringArray.coerce, desired_state: true
      property :max_value,
               Integer, coerce: ::Google::Sql::Property::Integer.coerce, desired_state: true
      property :min_value,
               Integer, coerce: ::Google::Sql::Property::Integer.coerce, desired_state: true
      property :f_label,
               String,
               coerce: ::Google::Sql::Property::String.coerce,
               name_property: true, desired_state: true
      property :requires_restart,
               kind_of: [TrueClass, FalseClass],
               coerce: ::Google::Sql::Property::Boolean.coerce, desired_state: true
      property :type, String, coerce: ::Google::Sql::Property::String.coerce, desired_state: true

      property :credential, String, desired_state: false, required: true
      property :project, String, desired_state: false, required: true

      action :create do
        Chef.deprecated(:generic,
                        ["gsql_flag has been deprecated.",
                         "Please use the Flag Name instead."
                        ].join(" "))

        fetch = fetch_wrapped_resource(@new_resource, 'sql#flag',
                                       'sql#flagsList',
                                       'items')
        if fetch.nil?
          converge_by "Creating gsql_flag[#{new_resource.name}]" do
            # TODO(nelsonjr): Show a list of variables to create
            # TODO(nelsonjr): Determine how to print green like update converge
            puts # making a newline until we find a better way TODO: find!
            compute_changes.each { |log| puts "    - #{log.strip}\n" }
            create_req = ::Google::Sql::Network::Post.new(
              collection(@new_resource), fetch_auth(@new_resource),
              'application/json', resource_to_request
            )
            return_if_object create_req.send, 'sql#flag'
          end
        else
          @current_resource = @new_resource.clone
          @current_resource.allowed_string_values =
            ::Google::Sql::Property::StringArray.api_parse(fetch['allowedStringValues'])
          @current_resource.applies_to =
            ::Google::Sql::Property::StringArray.api_parse(fetch['appliesTo'])
          @current_resource.max_value =
            ::Google::Sql::Property::Integer.api_parse(fetch['maxValue'])
          @current_resource.min_value =
            ::Google::Sql::Property::Integer.api_parse(fetch['minValue'])
          @current_resource.f_label = ::Google::Sql::Property::String.api_parse(fetch['name'])
          @current_resource.requires_restart =
            ::Google::Sql::Property::Boolean.api_parse(fetch['requiresRestart'])
          @current_resource.type = ::Google::Sql::Property::String.api_parse(fetch['type'])

          update
        end
      end

      action :delete do
        fetch = fetch_wrapped_resource(@new_resource, 'sql#flag', 'sql#flagsList', 'items')
        unless fetch.nil?
          converge_by "Deleting gsql_flag[#{new_resource.name}]" do
            delete_req = ::Google::Sql::Network::Delete.new(
              self_link(@new_resource), fetch_auth(@new_resource)
            )
            return_if_object delete_req.send, 'sql#flag'
          end
        end
      end

      # TODO(nelsonjr): Add actions :manage and :modify

      private

      action_class do
        def resource_to_request
          request = {
            kind: 'sql#flag',
            name: new_resource.f_label
          }.reject { |_, v| v.nil? }
          request.to_json
        end

        def unwrap_resource_filter(resource)
          self.class.unwrap_resource_filter(resource)
        end

        def self.unwrap_resource_filter(resource)
          {
            name: resource.f_label
          }
        end

        def update
          converge_if_changed do |_vars|
            # TODO(nelsonjr): Determine how to print indented like upd converge
            # TODO(nelsonjr): Check w/ Chef... can we print this in red?
            puts # making a newline until we find a better way TODO: find!
            compute_changes.each { |log| puts "    - #{log.strip}\n" }
            raise('Cloud SQL flag does not match expectations.')
          end
        end

        def self.resource_to_hash(resource)
          {
            project: resource.project,
            name: resource.f_label,
            kind: 'sql#flag',
            allowed_string_values: resource.allowed_string_values,
            applies_to: resource.applies_to,
            max_value: resource.max_value,
            min_value: resource.min_value,
            requires_restart: resource.requires_restart,
            type: resource.type
          }.reject { |_, v| v.nil? }
        end

        # Copied from Chef > Provider > #converge_if_changed
        def compute_changes
          properties = @new_resource.class.state_properties.map(&:name)
          properties = properties.map(&:to_sym)
          if current_resource
            compute_changes_for_existing_resource properties
          else
            compute_changes_for_new_resource properties
          end
        end

        # Collect the list of modified properties
        def compute_changes_for_existing_resource(properties)
          specified_properties = properties.select do |property|
            @new_resource.property_is_set?(property)
          end
          modified = specified_properties.reject do |p|
            @new_resource.send(p) == current_resource.send(p)
          end

          generate_pretty_green_text(modified)
        end

        def generate_pretty_green_text(modified)
          property_size = modified.map(&:size).max
          modified.map! do |p|
            properties_str = if @new_resource.sensitive
                               '(suppressed sensitive property)'
                             else
                               [
                                 @new_resource.send(p).inspect,
                                 "(was #{current_resource.send(p).inspect})"
                               ].join(' ')
                             end
            "  set #{p.to_s.ljust(property_size)} to #{properties_str}"
          end
        end

        # Write down any properties we are setting.
        def compute_changes_for_new_resource(properties)
          property_size = properties.map(&:size).max
          properties.map do |property|
            default = ' (default value)' \
              unless @new_resource.property_is_set?(property)
            next if @new_resource.send(property).nil?
            properties_str = if @new_resource.sensitive
                               '(suppressed sensitive property)'
                             else
                               @new_resource.send(property).inspect
                             end
            ["  set #{property.to_s.ljust(property_size)}",
             "to #{properties_str}#{default}"].join(' ')
          end.compact
        end

        def resource_to_query_predicate(resource)
          self.class.resource_to_query_predicate(resource)
        end

        def self.resource_to_query_predicate(resource)
          {
            name: resource.name
          }
        end

        def fetch_auth(resource)
          self.class.fetch_auth(resource)
        end

        def self.fetch_auth(resource)
          resource.resources("gauth_credential[#{resource.credential}]")
                  .authorization
        end

        def fetch_resource(resource, self_link, kind)
          self.class.fetch_resource(resource, self_link, kind)
        end

        def debug(message)
          Chef::Log.debug(message)
        end

        def self.collection(data)
          URI.join(
            'https://www.googleapis.com/sql/v1beta4/',
            expand_variables(
              'flags',
              data
            )
          )
        end

        def collection(data)
          self.class.collection(data)
        end

        def self.self_link(data)
          URI.join(
            'https://www.googleapis.com/sql/v1beta4/',
            expand_variables(
              'flags',
              data
            )
          )
        end

        def self_link(data)
          self.class.self_link(data)
        end

        # rubocop:disable Metrics/CyclomaticComplexity
        def self.return_if_object(response, kind)
          raise "Bad response: #{response.body}" \
            if response.is_a?(Net::HTTPBadRequest)
          raise "Bad response: #{response}" \
            unless response.is_a?(Net::HTTPResponse)
          return if response.is_a?(Net::HTTPNotFound)
          return if response.is_a?(Net::HTTPNoContent)
          result = JSON.parse(response.body)
          raise_if_errors result, %w[error errors], 'message'
          raise "Bad response: #{response}" unless response.is_a?(Net::HTTPOK)
          result
        end
        # rubocop:enable Metrics/CyclomaticComplexity

        def return_if_object(response, kind)
          self.class.return_if_object(response, kind)
        end

        def self.extract_variables(template)
          template.scan(/{{[^}]*}}/).map { |v| v.gsub(/{{([^}]*)}}/, '\1') }
                  .map(&:to_sym)
        end

        def self.expand_variables(template, var_data, extra_data = {})
          data = if var_data.class <= Hash
                   var_data.merge(extra_data)
                 else
                   resource_to_hash(var_data).merge(extra_data)
                 end
          extract_variables(template).each do |v|
            unless data.key?(v)
              raise "Missing variable :#{v} in #{data} on #{caller.join("\n")}}"
            end
            template.gsub!(/{{#{v}}}/, CGI.escape(data[v].to_s))
          end
          template
        end

        def self.fetch_resource(resource, self_link, kind)
          get_request = ::Google::Sql::Network::Get.new(
            self_link, fetch_auth(resource)
          )
          return_if_object get_request.send, kind
        end

        def fetch_wrapped_resource(resource, kind, wrap_kind, wrap_path)
          self.class.fetch_wrapped_resource(resource, kind, wrap_kind, wrap_path)
        end

        def self.fetch_wrapped_resource(resource, kind, wrap_kind, wrap_path)
          result = fetch_resource(resource, self_link(resource), wrap_kind)
          return if result.nil? || !result.key?(wrap_path)
          result = unwrap_resource(result[wrap_path], resource)
          return if result.nil?
          raise "Incorrect result: #{result['kind']} (expected #{kind})" \
            unless result['kind'] == kind
          result
        end

        def unwrap_resource(result, resource)
          self.class.unwrap_resource(result, resource)
        end

        def self.unwrap_resource(result, resource)
          query_predicate = unwrap_resource_filter(resource)
          matches = result.select do |entry|
            query_predicate.all? do |k, v|
              entry[k.id2name] == v
            end
          end
          raise "More than 1 result found: #{matches}" if matches.size > 1
          return if matches.empty?
          matches.first
        end

        def self.raise_if_errors(response, err_path, msg_field)
          errors = ::Google::HashUtils.navigate(response, err_path)
          raise_error(errors, msg_field) unless errors.nil?
        end

        def self.raise_error(errors, msg_field)
          raise IOError, ['Operation failed:',
                          errors.map { |e| e[msg_field] }.join(', ')].join(' ')
        end
      end
    end
    # rubocop:enable Metrics/ClassLength
  end
end
